/* (c) Joerg Schulenburg 2000-2017

 this is mpi/pthread-specific part of spin-package

 it is for division of matrix multiplication 
 in parallelizable block matrix multiplication
 
    b = H * a    (a,b vectors, H matrix)
    
    b[i] = sum_j H[i,j] * a[j]
    
 blocks can be multiplicated step by step on one system or
 they can be multiplicated on different processors/systems
 
 but you have to pay additional network/harddisk traffic
 
 if b_num is one you have only one block (similar to old version)
 
 - use mpi_n nodes and mpt_n (1..B_NUM) threads per node
 - variable block counts from 0 to pt_n-1 (B_NUM) in each node
 - set B_NUM or CFG_PTHREAD only if using OpenMP or pthreads
   it eats unused memory if bigger than 1 and no threading used (MPI-only)
   B_NUM can and should be set by Makefile.CFG_PTHREAD=... since 2016-05-18 
   B_NUM will be removed in future, default = CFG_PTHREAD = 1

 */

#ifndef _HBLOCK_H
#define _HBLOCK_H

#ifdef CONFIG_PTHREAD
/* 2017-02 mpiCC + icpc16 does not like it */
//#define _XOPEN_SOURCE 600  /* needed to define Barrier prototypes ??? */
#endif

#if defined(CFG_MPI) && defined(CONFIG_MPI)
#error only one MPI-switch must be defined outside /* 2016-05-18 debug */
#endif
#if defined(CFG_MPI) && !defined(CONFIG_MPI)
#define CONFIG_MPI CFG_MPI   /* 2016-05-18 shorter macro can be used too */
#endif
#if defined(CONFIG_MPI) && !defined(CFG_MPI)
#define CFG_MPI CONFIG_MPI   /* 2016-05-18 compatibility, ToDo: obsolete */
#endif

#ifdef CONFIG_MPI
#include <mpi.h>
#endif

#include "config.h"

/* maximum number of compute threads per mpi node, set by option -t<num>
 *  else set by OMP_NUM_THREADS (2016-05)
 * check nodes for: grep -c ^processor /proc/cpuinfo
 *   or: pbsnodes|grep "^\ *np\ =" |sort -u -k 3n|head -1|awk '{print $3}'
 *   or: sinfo -h -o %c
 */
#ifndef B_NUM  /* see remaks at top */
#if     defined(CONFIG_PTHREAD) && CONFIG_PTHREAD>1
#define B_NUM CONFIG_PTHREAD /* 2016-05-18 can be set by configure/Makefile now */
#elif   defined(CONFIG_PTHREAD) && !defined(CONFIG_MPI)
#define B_NUM 32 /* may be 1000 or more on massive SMP systems */
#elif   defined(CONFIG_PTHREAD) &&  defined(CONFIG_MPI) /* 2016-04-22 ~! */
#define B_NUM 8  /* spare memory on small nodes */
#else
#define B_NUM 1  /* set to minimum to spare memory */
#endif
#endif

/* ToDo2016 autodetect MPI_MAX_NUM by:
 *   pbsnodes | grep -c ^[a-z]  # comp2
 * x pbsnodes|grep "^\ *np\ =" |sort -u -k 3n|head -1|awk '{print $3}'
 *   sinfo -h -o %D             # t100,kautz
 * x sinfo -h -o %c             #  if B_NUM=1
 * 2017: 16K tasks (1024nodes *16cores)
 * 2017-01-18 enlarge from 8K to 64K/B_NUM fits to int16 (*4B=256KB), bigger default
 *   used for Alltoallv-tables and l1_ofs
 */
#ifndef MPI_MAX_NUM   /* ToDo2016: use CFG_MPI > 1 */
#define MPI_MAX_NUM (8192+32*1024/(B_NUM))  /* max number MPI tasks for MPI_buffers */
#endif

/* #include "defs.h" */

extern int mpi_n, mpi_i;	/* number of MPI nodes + nodenr.         */
extern int pt_n;                /* number of threads per node (1..B_NUM) */
extern int pt_n_bits, mpi_n_bits; /* bits needed to store pt_i and mpi_i */
/* pointer for blocked and locally stored vectors (MPI!) */
/* see b_ini: node_ofs=mpi_ofs[mpi_i]; b_ofs[i]=0..node_n1 */
extern unsigned long node_ofs, node_n1;	/* part of space on this node */
/* pointer for blocked and locally stored vectors (NUMA!) [mpi_n*pt_n] */
// extern unsigned long *b_ofs;	/* start index 0..n1-1 of block (obsolete) */
// extern unsigned long *b_len;	/* block len (obsolete) */
// extern unsigned long max_b_len; /* maximum b_len to ensure mpi balance */

/* this offset definitions are new in v2.36, and replace old pt_n*mpi_n
 *    which have required aditional operations in the inner loop 
 * start index 0..n1 of vector in thread[0..pt_n] for this mpi node */
extern unsigned long b_ofs[B_NUM+1];
/* start index 0..n1 of vector in mpi_node[0..mpi_n] */
extern unsigned long mpi_ofs[MPI_MAX_NUM+1];
/* absolute ofs: mpi_ofs[mpi_i]+b_ofs[pthread_i]
 * block length: mpi_ofs[mpi_n]==n1
 *               b_ofs[pt_n]=node_n1
 *                          =mpi_ofs[mpi_i+1]-mpi_ofs[mpi_i]
 */
//extern unsigned long max_b_len; /* maximum block length ensure mpi-balance */

/* this is for stored files, ToDo: move to h_file */
extern unsigned long b_hnz[B_NUM];	/* NumNonZero-H block entries */
/* this is for partly stored files, ToDo: remove or replace */
extern unsigned long b_snz[B_NUM];	/* stored NonZero-H block entries */
//extern unsigned long b_sn1[B_NUM];	/* stored block lines, new _ah_blks */
extern int b_bits;			/* b_len[0..n-2]=1l<<b_bits */
extern unsigned long b_mask;		/* ~((~0)<<b_bits) */
extern unsigned long len_full_block;   /* number of elements per block */

/* initialization, n1 is whole matrix size */
int b_ini(unsigned long n1,int vvv);
void b_free(); /* opposit of b_ini */

#include "h_file.h"
extern struct thxy b_hbuf[B_NUM]; /* buffer for faster access, [HBLen] */
extern int          b_bnz[B_NUM]; /* num valid nz-data in buffer bufh */  
extern int          b_bn1[B_NUM]; /* num valid lines in buffer bufh */  

/* 2015-12-31 testwise, ToDo: pthread-alternative */
#ifdef _OPENMP
#include <omp.h>
#else
#define omp_get_max_threads() 1
#define omp_get_thread_num() 0
#endif

/* 2017-02-02 moved here from spins.c */
#ifdef CONFIG_PTHREAD   /* ToDo: movto mthreads.h, or write new OpenMP-code */
#ifndef _OPENMP /* OPENMP is replacing PTHREAD */
#include <pthread.h>
extern pthread_attr_t pt_attr;  /* SunOS has SCOPE_PROCESS as default, we need SCOPE_SYSTEM */
extern pthread_cond_t pt_cond1;  /* conditions master -> childs */
extern pthread_cond_t pt_cond2;  /* conditions childs -> master */
extern pthread_mutex_t pt_mutex1; /* mutex */
extern pthread_mutex_t pt_mutex2; /* mutex */
/* MPI code is in not thread safe in general, we have to care about it */
extern pthread_mutex_t token_ring[B_NUM*4]; /* 4 rings of double(quad) length */
extern pthread_barrier_t pt_barrier; /* 2017-01-18 for new better mixed code (MPI+PT) */
#else /* OPENMP-replacements? PTHREAD-dummys! */
#define pthread_mutex_lock(x)
#define pthread_mutex_unlock(x)
#define pthread_barrier_wait(x) omp_barrier_wait()
extern volatile int omp_token;
extern volatile int omp_token2;
static inline int omp_barrier_wait() {
#pragma omp atomic   // omp-3.0 only? gcc412.omp2.5=OK? need flush? ToDo17
    omp_token++; // OMP_BARRIER cant be used, because of our master thread?
    while (omp_token%pt_n!=0); /* usleep? */
    omp_token2++; // OMP_BARRIER cant be used, because of our master thread?
    while (omp_token2%pt_n!=0); /* usleep? */
    return 0;
}
#define pthread_cond_broadcast(x)
#endif /* ~_OPENMP */
#else  /* ~PTHREAD dummy definitions to eleminate ifdef constructions */
#define pthread_mutex_lock(x)
#define pthread_mutex_unlock(x)
#define pthread_barrier_wait(x)
#define pthread_cond_broadcast(x)         
#endif

#ifndef CONFIG_MPI  /* define MPI dummy-functions to reduce ifdef-lines */
#define MPI_Barrier(MPI_COMM_WORLD)  /* dummy function for ~MPI */
#define MPI_Bcast(abuf,cnt,MPI_INT,root,MPI_COMM_WORLD) /* 2012-05 */
// #define MPI_Bcast(x1,x2,x3,x4,x5) /* dummy 2012-05 */
#define MPI_Abort(MPI_COMM_WORLD,x2)  exit(x2) /* dummy*/
// MPI_Allreduce(&add, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#define MPI_Allreduce(add, sum, cnt, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD)\
   { int i; for (i=0;i<(cnt);i++) (sum)[i]=(add)[i]; } /* 2016-04-15 */
#endif // ~CONFIG_MPI

#endif   /* ifndef _HBLOCK_H */
